# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from swh.core.api.classes import stream_results
from swh.loader.core.loader import BaseLoader
from swh.loader.metadata.base import BaseMetadataFetcher
from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    Origin,
    RawExtrinsicMetadata,
)

ORIGIN = Origin(url="some-url")

METADATA_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.FORGE, url="http://example.org/"
)
REMD = RawExtrinsicMetadata(
    target=ORIGIN.swhid(),
    discovery_date=datetime.datetime.now(tz=datetime.timezone.utc),
    authority=METADATA_AUTHORITY,
    fetcher=MetadataFetcher(
        name="test fetcher",
        version="0.0.1",
    ),
    format="test-format",
    metadata=b'{"foo": "bar"}',
)


class DummyLoader(BaseLoader):
    """Base Loader to overload and simplify the base class (technical: to avoid repetition
    in other *Loader classes)"""

    visit_type = "git"

    def __init__(self, storage, *args, **kwargs):
        super().__init__(storage, *args, **kwargs)

    def cleanup(self):
        pass

    def prepare(self, *args, **kwargs):
        pass

    def fetch_data(self):
        pass

    def get_snapshot_id(self):
        return None

    def store_data(self):
        pass


class DummyMetadataFetcher(BaseMetadataFetcher):
    SUPPORTED_LISTERS = {"fake-lister"}
    FETCHER_NAME = "dummy"

    def __init__(self, origin, credentials, lister_name, lister_instance_name):
        pass

    def get_origin_metadata(self):
        return [REMD]

    def get_parent_origins(self):
        return []


def test_load(swh_storage, mocker):
    mocker.patch(
        "swh.loader.core.metadata_fetchers._fetchers",
        return_value=[DummyMetadataFetcher],
    )

    loader = DummyLoader(
        storage=swh_storage,
        origin_url=ORIGIN.url,
        lister_name="fake-lister",
        lister_instance_name="",
    )
    loader.load()

    assert swh_storage.raw_extrinsic_metadata_get(
        ORIGIN.swhid(), METADATA_AUTHORITY
    ).results == [REMD]


def test_load_unknown_lister(swh_storage, mocker):
    mocker.patch(
        "swh.loader.core.metadata_fetchers._fetchers",
        return_value=[DummyMetadataFetcher],
    )

    loader = DummyLoader(
        storage=swh_storage,
        origin_url=ORIGIN.url,
        lister_name="other-lister",
        lister_instance_name="",
    )
    loader.load()

    assert (
        list(
            stream_results(
                swh_storage.raw_extrinsic_metadata_get,
                ORIGIN.swhid(),
                METADATA_AUTHORITY,
            )
        )
        == []
    )



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import dataclasses
import datetime
from typing import Any, Dict, Iterable, List, Type, TypeVar, cast
import uuid

from swh.core.api.classes import stream_results
from swh.core.statsd import Statsd
from swh.loader.core.metadata_fetchers import CredentialsType, get_fetchers_for_lister
from swh.loader.metadata.base import BaseMetadataFetcher
from swh.model.model import (
    MetadataAuthority,
    MetadataFetcher,
    Origin,
    RawExtrinsicMetadata,
)
from swh.scheduler.interface import ListedOrigin, Lister, SchedulerInterface
from swh.storage.interface import StorageInterface


def _now() -> datetime.datetime:
    # Used by tests for mocking
    return datetime.datetime.now(tz=datetime.timezone.utc)


_TItem = TypeVar("_TItem")


@dataclasses.dataclass
class JournalClient:
    scheduler: SchedulerInterface
    storage: StorageInterface
    metadata_fetcher_credentials: CredentialsType
    reload_after_days: int

    def __post_init__(self):
        self._listers = {}
        self._added_fetchers = set()
        self._added_authorities = set()
        self.statsd = Statsd(namespace="swh_loader_metadata_journal_client")

    def statsd_timed(self, name: str, tags: Dict[str, Any] = {}):
        """
        Wrapper for :meth:`swh.core.statsd.Statsd.timed`, which uses the standard
        metric name and tag.
        """
        return self.statsd.timed(
            "operation_duration_seconds", tags={"operation": name, **tags}
        )

    def statsd_timing(self, name: str, value: float, tags: Dict[str, Any] = {}) -> None:
        """
        Wrapper for :meth:`swh.core.statsd.Statsd.timing`, which uses the standard
        metric name and tags for loaders.
        """
        self.statsd.timing(
            "operation_duration_seconds", value, tags={"operation": name, **tags}
        )

    def _get_lister(self, lister_id: uuid.UUID) -> Lister:
        if lister_id not in self._listers:
            with self.statsd_timed("get_listers_by_id"):
                (lister,) = self.scheduler.get_listers_by_id([str(lister_id)])
            self._listers[lister.id] = lister
        return self._listers[lister_id]

    def _add_metadata_fetchers(self, fetchers: Iterable[MetadataFetcher]) -> None:
        for fetcher in fetchers:
            if fetcher not in self._added_fetchers:
                self.storage.metadata_fetcher_add([fetcher])
                self._added_fetchers.add(fetcher)

    def _add_metadata_authorities(
        self, authorities: Iterable[MetadataAuthority]
    ) -> None:
        for authority in authorities:
            if authority not in self._added_authorities:
                self.storage.metadata_authority_add([authority])
                self._added_authorities.add(authority)

    def process_journal_objects(self, messages: Dict[str, List[Dict]]) -> None:
        """Loads metadata for origins not recently loaded:

        1. reads messages from the origin journal topic
        2. queries the scheduler for a list of listers that produced this origin
           (to guess what type of forge it is)
        3. if it is a forge we can get extrinsic metadata from, check if we got any
           recently, using the storage
        4. if not, trigger a metadata load
        """

        assert set(messages) == {"origin"}, f"Unexpected message types: {set(messages)}"

        for origin in messages["origin"]:

            for listed_origin in stream_results(
                self.statsd_timed("get_listed_origins")(
                    self.scheduler.get_listed_origins
                ),
                url=origin["url"],
            ):
                self._process_listed_origin(listed_origin)

        self.storage.flush()

    def _process_listed_origin(
        self,
        listed_origin: ListedOrigin,
    ) -> List[RawExtrinsicMetadata]:
        origin = Origin(url=listed_origin.url)

        lister = self._get_lister(listed_origin.lister_id)

        tags = {
            "lister": lister.name,
            "lister_instance": lister.instance_name,
        }

        fetcher_classes = cast(
            List[Type[BaseMetadataFetcher]], get_fetchers_for_lister(lister.name)
        )
        self.statsd.histogram("metadata_fetchers", len(fetcher_classes), tags=tags)

        now = _now()

        metadata: List[RawExtrinsicMetadata] = []

        for cls in fetcher_classes:
            tags["fetcher"] = cls.FETCHER_NAME

            metadata_fetcher = cls(
                origin=origin,
                lister_name=lister.name,
                lister_instance_name=lister.instance_name,
                credentials=self.metadata_fetcher_credentials,
            )

            with self.statsd_timed("raw_extrinsic_metadata_get"):
                last_metadata = self.storage.raw_extrinsic_metadata_get(
                    target=origin.swhid(),
                    authority=metadata_fetcher.metadata_authority(),
                    after=now - datetime.timedelta(days=self.reload_after_days),
                    limit=1,
                )

            if last_metadata.results:
                # We already have recent metadata; don't load it again.
                self.statsd.increment(
                    "metadata_items_fetched_total", len(last_metadata.results), tags=tags
                )

                continue

            with self.statsd_timed("get_origin_metadata", tags=tags):
                metadata = list(metadata_fetcher.get_origin_metadata())

            self.statsd.increment("new_metadata_items", len(metadata), tags=tags)

            with self.statsd_timed("metadata_fetcher_add"):
                self._add_metadata_fetchers({m.fetcher for m in metadata})

            with self.statsd_timed("metadata_authority_add"):
                self._add_metadata_authorities({m.authority for m in metadata})

            with self.statsd_timed("raw_extrinsic_metadata_add"):
                self.storage.raw_extrinsic_metadata_add(metadata)

        return metadata



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Metadata fetcher for GitHub."""

import json
import re
from typing import List, Optional, Tuple
import urllib.parse

from swh.lister.github.utils import GitHubSession
from swh.model.model import Origin

from . import USER_AGENT
from .base import BaseMetadataFetcher, InvalidOrigin

HTTP_ACCEPT = "application/vnd.github.v3+json"
"""HTTP header sent on all API requests to GitHub."""

# The format is defined by a well-understood MIME type; we might as well use that.
METADATA_FORMAT = HTTP_ACCEPT
"""Value of the ``format`` field of produced
:class:`swh.model.model.RawExtrinsicMetadata` objects."""

_API_URL = "https://api.github.com/repos{path}"


class GitHubMetadataFetcher(BaseMetadataFetcher):
    FETCHER_NAME = "github"
    SUPPORTED_LISTERS = {"github"}

    _github_session: Optional[GitHubSession] = None

    def github_session(self) -> GitHubSession:
        if self._github_session is None:
            self._github_session = GitHubSession(
                user_agent=USER_AGENT, credentials=self.credentials
            )
        return self._github_session

    def _check_origin(self):
        (scheme, netloc, path, query, fragment) = urllib.parse.urlsplit(self.origin.url)
        if netloc != "github.com":
            # TODO: relax this check when we support self-hosted GitHub instances
            raise InvalidOrigin(f"netloc should be 'github.com', not '{netloc}'")

        if scheme != "https" or not re.match(r"/[^\s/]+/[^\s/]+", path):
            raise InvalidOrigin(f"Unsupported github.com URL: {self.origin.url}")

        if query != "" or fragment != "":
            raise InvalidOrigin(
                f"Unexpected end query or fragment in github.com URL: {self.origin.url}"
            )

    def _get_origin_metadata_bytes(self) -> List[Tuple[str, bytes]]:
        (scheme, netloc, path, query, fragment) = urllib.parse.urlsplit(self.origin.url)
        response = self.github_session().request(_API_URL.format(path=path))
        if response.status_code != 200:
            # TODO: retry
            return []

        metadata_bytes = response.content

        # TODO?: strip API hyperlinks from metadata_bytes to save space?
        # They take 10KB for every repo, or 1KB when compressed by the database server.
        # This means processing metadata_bytes and changing the format, instead of
        # archiving verbatim, though.

        return [(METADATA_FORMAT, metadata_bytes)]

    def get_parent_origins(self) -> List[Origin]:
        parents = []
        for metadata in self.get_origin_metadata():
            if metadata.format != METADATA_FORMAT:
                continue
            data = json.loads(metadata.metadata)
            parent = data.get("parent")
            source = data.get("source")
            if parent is not None:
                parents.append(Origin(url=parent["html_url"]))
                if source is not None and source["html_url"] != parent["html_url"]:
                    parents.append(Origin(url=source["html_url"]))

        return parents
